
;;;======================================================
;;;   Tic Tac Toe Sample Problem
;;;
;;;     This is a rather simple tic tac toe expert
;;;     system. It will usually win if possible, but
;;;     won't always stop you from winning.
;;;
;;;     To execute, merely load, reset and run.
;;;======================================================

(deffacts initial-state
  (square 1 1 e corner)
  (square 2 1 e side)
  (square 3 1 e corner)
  (square 1 2 e side)
  (square 2 2 e center)
  (square 3 2 e side)
  (square 1 3 e corner)
  (square 2 3 e side)
  (square 3 3 e corner))

;;;******************************************
;;; evaluate board for blocks, wins, and ties
;;;******************************************

(defrule evaluate-potential-win ""
  (declare (salience 10))
  (evaluate board for ?)
  (square ?row1 ?col1 ?pl&~e ?)
  (square ?row2 ?col2 e ?)
  (test (|| (!= ?row1 ?row2) (!= ?col1 ?col2)))
  (square ?row3 ?col3 e ?type)
  (test (&& (||  (!= ?row1 ?row3) (!= ?col1 ?col3))
	            (||  (!= ?row2 ?row3) (!= ?col2 ?col3))))
  (test (||  (&& (= ?row1 ?row2)
		                (= ?row1 ?row3))
	            (&& (= ?col1 ?col2)
		                (= ?col1 ?col3))
	            (&& (= (+ ?row1 ?col1) (+ ?row2 ?col2))
		                (= (+ ?row1 ?col1) (+ ?row3 ?col3)))
	            (&& (= (- ?row1 ?col1) (- ?row2 ?col2))
		                (= (- ?row1 ?col1) (- ?row3 ?col3)))))
   =>
   (assert (potential-win ?row3 ?col3 ?type ?pl)))

(defrule evaluate-immediate-block-or-win ""
  (declare (salience 10))
  (evaluate board for ?)
  (square ?row1 ?col1 ?pl&~e ?)
  (square ?row2 ?col2 ?pl ?)
  (test (|| (!= ?row1 ?row2) (!= ?col1 ?col2)))
  (square ?row3 ?col3 e ?type)
  (test (&& (||  (!= ?row1 ?row3) (!= ?col1 ?col3))
	            (||  (!= ?row2 ?row3) (!= ?col2 ?col3))))
  (test (||  (&& (= ?row1 ?row2)
		                (= ?row1 ?row3))
	            (&& (= ?col1 ?col2)
		                (= ?col1 ?col3))
	            (&& (= (+ ?row1 ?col1) (+ ?row2 ?col2))
		                (= (+ ?row1 ?col1) (+ ?row3 ?col3)))
	            (&& (= (- ?row1 ?col1) (- ?row2 ?col2))
		                (= (- ?row1 ?col1) (- ?row3 ?col3)))))
   =>
   (assert (immediate-block-or-win ?row3 ?col3 ?type ?pl)))

(defrule tie-game ""
  (declare (salience 10))
  ?phase <- (evaluate board for ?)
  (not (square ? ? e ?))
  =>
  (retract ?phase)
  (printout tab "The game has ended in a tie." crlf))

;;;*****************************
;;; rules for the computers move
;;;*****************************

(defrule move-computer-to-win
  (declare (salience 10))
  ?phase <- (computer move)
  (computer is ?cs)
  ?dummy <- (immediate-block-or-win ?row ?col ?type ?cs)
  ?square <- (square ?row ?col ? ?)
  =>
  (assert (display board))
  (retract ?dummy ?phase ?square)
  (assert (square ?row ?col ?cs ?type))
  (printout tab "Computer moves to win!" crlf))

(defrule move-computer-to-block ""
  (declare (salience 10))
  ?phase <- (computer move)
  (computer is ?cs)
  (not (immediate-block-or-win ? ? ? ?cs))
  (human is ?hs)
  ?dummy <- (immediate-block-or-win ?row ?col ?type ?hs)
  ?square <- (square ?row ?col ? ?)
  =>
  (retract ?dummy ?phase ?square)
  (assert (display board))
  (assert (square ?row ?col ?cs ?type))
  (assert (next-move human)))

(defrule move-to-center ""
  (declare (salience 10))
  ?phase <- (computer move)
  (computer is ?cs)
  (not (immediate-block-or-win ? ? ? ?))
  ?square <- (square ? ? e center)
  =>
  (retract ?phase ?square)
  (assert (display board))
  (assert (square 2 2 ?cs center))
  (assert (next-move human)))

(defrule move-to-corner ""
  (declare (salience 10))
  (computer is ?cs)
  ?phase <- (computer move)
  (not (immediate-block-or-win ? ? ? ?)) 
  ?square <- (square ?row ?col e corner)
  (not (square ? ? e center))
  =>
  (retract ?phase ?square)
  (assert (display board))
  (assert (square ?row ?col ?cs corner))
  (assert (next-move human)))

(defrule move-to-side ""
  (declare (salience 10))
  ?phase <- (computer move)
  (computer is ?cs)
  (not (immediate-block-or-win ? ? ? ?)) 
  ?square <- (square ?row ?col e side) 
  (not (square ? ? e corner))
  (not (square ? ? e center))
  =>
  (retract ?phase ?square)
  (assert (display board))
  (assert (square ?row ?col ?cs side))
  (assert (next-move human)))

;;;*******************************************
;;; Get human move and determine validity
;;;*******************************************

(defrule human-input ""
  (declare (salience 10))
  (human move)
  =>
  (printout tab "Column? ")
  (bind ?col (read))
  (printout tab "Row? ")
  (bind ?row (read))
  (printout tab)
  (assert (human-to ?col ?row)))

(defrule valid-human-move ""
  (declare (salience 10))
  ?phase <- (human move)
  (human is ?hs)
  ?move  <- (human-to ?row ?col)
  ?square <- (square ?row ?col e ?type)
  =>
  (assert (display board))
  (retract ?phase ?move ?square)
  (assert (square ?row ?col ?hs ?type))
  (assert (valid-move ?row ?col)))

(defrule human-winning-move ""
  (declare (salience 10))
  ?move <- (valid-move ?row ?col)
  (human is ?hs)
  ?info <- (immediate-block-or-win ?row ?col ?type ?hs)
  =>
  (retract ?move ?info)
  (printout tab "You have made the winning move!" crlf))

(defrule human-blocking-move ""
  (declare (salience 10))
  ?move <- (valid-move ?row ?col)
  (computer is ?cs)
  ?info <- (immediate-block-or-win ? ? ?type ?cs)
  =>
  (retract ?info ?move)
  (assert (next-move computer)))

(defrule no-block-or-win-for-human ""
  (declare (salience 10))
  ?move <- (valid-move ?row ?col)
  (not (immediate-block-or-win ? ? ? ?))
  =>
  (retract ?move)
  (assert (next-move computer)))

(defrule invalid-human-move ""
  (declare (salience 10))
  ?phase <- (human move)
  ?move  <- (human-to ?row ?col)
  (square ?row ?col ~e ?)
  =>
  (printout tab "Invalid move. Try another." crlf)
  (retract ?phase ?move)
  (assert (human move)))

;;;******************************
;;; cleanup rules
;;;******************************

(defrule cleanup-1 ""
  (declare (salience 10))
  (cleanup for ?)
  ?trash <- (immediate-block-or-win ? ? ? ?)
  =>
  (retract ?trash))

(defrule cleanup-2 ""
  (declare (salience 10))
  (cleanup for ?)
  ?trash <- (potential-win ? ? ? ?)
  =>
  (retract ?trash))

;;;******************************
;;; phase control rules
;;;******************************

(defrule setup ""
  (declare (salience 10))
  (initial-fact)
  =>
  (printout tab "Should the computer move first? ")
  (bind ?resp (read))
  (printout crlf)
  (if (|| (eq ?resp yes) (eq ?resp y))
      then
      (assert (computer is o))
      (assert (human is x))
      (assert (evaluate board for computer))
      else
      (assert (computer is x))
      (assert (human is o))
      (assert (evaluate board for human))))

(defrule cleanup-for-player ""
  ?phase <- (next-move ?player)
  =>
  (retract ?phase)
  (assert (cleanup for ?player)))

(defrule switch-to-evaluate ""
  ?phase <- (cleanup for ?player)
  =>
  (retract ?phase)
  (assert (evaluate board for ?player)))

(defrule switch-to-move ""
  ?phase <- (evaluate board for ?player)
  =>
  (retract ?phase)
  (printout tab "*********" crlf)
  (printout tab ?player " move" crlf)
  (printout tab "*********" crlf)
  (assert (?player move)))

(defrule board ""
  (declare (salience 10))	
  ?phase <- (display board)
  (square 1 1 ?p1-1 ?)
  (square 2 1 ?p2-1 ?)
  (square 3 1 ?p3-1 ?)
  (square 1 2 ?p1-2 ?)
  (square 2 2 ?p2-2 ?)
  (square 3 2 ?p3-2 ?)
  (square 1 3 ?p1-3 ?)
  (square 2 3 ?p2-3 ?)
  (square 3 3 ?p3-3 ?)
  =>
  (retract ?phase)
  (printout crlf "Board" crlf)
  (printout tab " " ?p1-1 "|" ?p2-1 "|" ?p3-1 crlf)
  (printout tab " -----" crlf)
  (printout tab " " ?p1-2 "|" ?p2-2 "|" ?p3-2 crlf)
  (printout tab " -----" crlf)
  (printout tab " " ?p1-3 "|" ?p2-3 "|" ?p3-3 crlf crlf))

